import copy
import enum
from urllib.parse import urljoin

import requests

from postgrestutils.client.utils import datetime_parser

REPR_OUTPUT_SIZE = 20

Count = enum.Enum('Count', (('NONE', None), ('EXACT', 'exact')))


class PostgrestClient:
    def __init__(self, base_uri, token=None):
        self.session = requests.Session()
        self.configure(token, base_uri=base_uri)
        self.session.headers['Accept'] = 'application/json'

    def configure(self, token, base_uri=None):
        if base_uri is not None:
            self.base_uri = base_uri
        if token:
            self.session.headers['Authorization'] = 'Bearer {}'.format(token)

    def get(self, endpoint, singular=False, parse_dt=True, count=Count.NONE, **kwargs):
        """
        :param endpoint: specifies which endpoint to request
        :param singular: if True returns a JSON object rather than a list (406
        when multiple results are returned)
        :param parse_dt: if True parses datetime strings as returned by
        PostgREST to python datetime objects
        :param count: counting strategy as explained in the README
        :param kwargs: pass kwargs directly to requests's .get() method
        :return: single element as dict (singular=True), lazy python object for
        multiple elements or raises HTTPError
        """
        if singular:  # immediately evaluate and return result
            res = LazyPostgrestJsonResult(self, endpoint, singular, parse_dt, count, **kwargs)
            res._fetch_all()  # will raise requests.HTTPError if no or multiple elements are returned
            return res._result_cache[0]
        return LazyPostgrestJsonResult(self, endpoint, singular, parse_dt, count, **kwargs)


class LazyPostgrestJsonResult:
    def __init__(self, client, endpoint, singular, parse_dt, count, **kwargs):
        self._len_cache = None
        self._result_cache = None

        self.client = client
        self.endpoint = endpoint
        self.singular = singular
        self.parse_dt = parse_dt
        self.count = count
        self.request_kwargs = kwargs

    def __repr__(self):
        data = list(self[:REPR_OUTPUT_SIZE + 1])
        if len(data) > REPR_OUTPUT_SIZE:
            data[-1] = "...(remaining elements truncated)..."
        return '<{} {}>'.format(self.__class__.__name__, data)

    def __len__(self):
        if self.count != Count.NONE:
            self._fetch_len()
        else:
            self._fetch_all()
        return self._len_cache

    def _fetch_len(self):
        if self._len_cache is None:
            request_kwargs = copy.deepcopy(self.request_kwargs)
            request_kwargs.setdefault('headers', dict())['Prefer'] = 'count={}'.format(self.count.value)
            request_kwargs['headers']['Range-Unit'] = 'items'
            # Have to request something so just fetch the first item
            request_kwargs['headers']['Range'] = '0-0'
            if self.singular:
                request_kwargs['headers']['Accept'] = 'application/vnd.pgrst.object+json'

            resp = self.client.session.get(urljoin(self.client.base_uri, self.endpoint), **request_kwargs)

            count = int(resp.headers['Content-Range'].split('/')[-1])
            self._len_cache = count

            # If the request yields only one element anyway, might as well cache
            # it. When using singular=False, count=Count.EXACT and the result
            # is a single element this saves an API request in cases where
            # len() is called before using the result.
            if count == 1:
                self._result_cache = self._parse_response(resp)

    def _fetch_all(self):
        if self._result_cache is None:
            request_kwargs = copy.deepcopy(self.request_kwargs)

            if self.singular:
                request_kwargs.setdefault('headers', dict())['Accept'] = 'application/vnd.pgrst.object+json'

            resp = self.client.session.get(urljoin(self.client.base_uri, self.endpoint), **request_kwargs)
            self._result_cache = self._parse_response(resp)

            # fetched all elements anyway, caching their length is very cheap
            self._len_cache = len(self._result_cache)

    def _fetch_some(self, range):
        request_kwargs = copy.deepcopy(self.request_kwargs)
        request_kwargs.setdefault('headers', dict())['Range-Unit'] = 'items'
        request_kwargs['headers']['Range'] = range

        if self.singular:
            request_kwargs['headers']['Accept'] = 'application/vnd.pgrst.object+json'

        resp = self.client.session.get(urljoin(self.client.base_uri, self.endpoint), **request_kwargs)
        return self._parse_response(resp)

    def _parse_response(self, resp):
        try:
            resp.raise_for_status()
        except requests.HTTPError as e:
            raise type(e)(resp.status_code, resp.reason, resp.text)

        if self.parse_dt:
            json_result = resp.json(object_hook=datetime_parser)
        else:
            json_result = resp.json()
        # always return a list even if it contains a single element only
        return [json_result] if self.singular else json_result

    def __getitem__(self, key):
        if not isinstance(key, (int, slice)):
            raise TypeError(
                "{self.__class__.__name__} indices must be integers or slices, not {key.__class__.__name__}".format(
                    self=self,
                    key=key
                )
            )
        if ((isinstance(key, int) and key < 0) or
           (isinstance(key, slice) and ((key.start is not None and key.start < 0) or
                                        (key.stop is not None and key.stop < 0)))):
            raise ValueError("{self.__class__.__name__} does not support negative indexing".format(self=self))
        if isinstance(key, slice) and key.step is not None:
            raise ValueError("{self.__class__.__name__} does not support stepping".format(self=self))

        if self._result_cache is not None:
            return self._result_cache[key]

        if isinstance(key, slice):
            range = '{start}-{stop}'.format(
                start=key.start or 0,
                stop=key.stop and key.stop - 1 or ''
            )
        else:
            range = '{0}-{0}'.format(key)
        return self._fetch_some(range)

    def refresh_from_pgrest(self):
        """Lazily refresh data from PostgREST"""
        self._result_cache = None
        self._len_cache = None

    def __iter__(self):
        self._fetch_all()
        return iter(self._result_cache)
