import re
from datetime import datetime, timedelta, timezone

DJANGO = False
try:
    from django.utils import dateparse, timezone as django_tz
    DJANGO = True
except ImportError:
    pass

# this regex matches postgres' JSON formatting for timestamps
JSON_TIMESTAMP_REGEX = re.compile(r"(?P<year>\d{4})-"
                                  r"(?P<month>\d{2})-"
                                  r"(?P<day>\d{2})"
                                  r"T(?P<hour>\d{2}):"
                                  r"(?P<minute>\d{2}):"
                                  r"(?P<second>\d{2})\."
                                  r"(?P<microsecond>\d{1,6})"
                                  r"((?P<offsetsign>[+-])"
                                  r"(?P<offsethours>\d{2}):"
                                  r"(?P<offsetminutes>\d{2}))?$")


def clean_parts(parts):
    cleaned = {}
    for key, value in parts.items():
        if value:
            if "offset" not in key:
                cleaned[key] = int(value)
            else:
                cleaned[key] = value
    return cleaned


def datetime_parser(json_dict):
    for key, value in json_dict.items():
        if isinstance(value, str):
            if DJANGO:
                try:
                    parsed_dt = dateparse.parse_datetime(value)
                    if parsed_dt:
                        if django_tz.is_naive(parsed_dt):
                            parsed_dt = django_tz.make_aware(parsed_dt)
                        json_dict[key] = parsed_dt
                except ValueError:
                    pass  # not a datetime, leave the string as it is
            else:
                match = JSON_TIMESTAMP_REGEX.match(value)
                if match:
                    parts = clean_parts(match.groupdict())
                    if parts.get('offsetsign') and parts.get('offsethours') and parts.get('offsetminutes'):
                        sign = -1 if parts.pop('offsetsign', '+') == '-' else 1
                        tz = timezone(
                            offset=sign * timedelta(
                                hours=int(parts.pop('offsethours')),
                                minutes=int(parts.pop('offsetminutes'))
                            )
                        )
                        parsed_dt = datetime(**parts).replace(tzinfo=tz).astimezone()
                    else:
                        # naive datetime so we assume local time
                        local_tz = datetime.now(timezone.utc).astimezone().tzinfo
                        parsed_dt = datetime(**parts).replace(tzinfo=local_tz)
                    json_dict[key] = parsed_dt
    return json_dict
